import numpy as np
from numpy import cos,sin,pi
import os

from fc_tools.others import func2str,isModuleFound,is_function
from fc_oogmsh import gmsh
from fc_simesh.siMesh import siMesh
from fc_vfemp1.operators import Loperator
from fc_vfemp1.BVP import BVP,PDE

from fc_vfemp1_surface.sys import get_geo
import fc_vfemp1_surface.lib as slib
from fc_vfemp1_surface.examples import sBVPsamples


def run(**kwargs):
  N=kwargs.pop('N',30)
  verbose=kwargs.pop('verbose',1)
  u=kwargs.pop('u',sBVPsamples.uxyz04)
  f=kwargs.pop('f',sBVPsamples.fxyz04)
  Fmap=kwargs.pop('Fmap',sBVPsamples.Fmapxyz04)
  gradSu=kwargs.pop('gradSu',sBVPsamples.gradSuxyz04)
  c=kwargs.pop('c',[lambda x1,x2,x3: cos(x1), lambda x1,x2,x3: sin(x2), lambda x1,x2,x3: 2+x1*x2*x3])
  a0=kwargs.pop('a0',lambda x,y,z: 1+x**2)
  alpha=kwargs.pop('alpha',lambda x,y,z: 1+z**2)
  graphics=kwargs.pop('graphics',1)
  geofile='halfsphere4surf'
  (geodir,geofile)=get_geo(3,2,geofile)
  geoFile=geodir+os.sep+geofile+'.geo'
  if verbose>0:
    slib.print_separator()
    slib.PrintCopyright()
    slib.print_separator()
    print('      Surface BVP on unit sphere maps by function Fmap')
    print('      Mesh generated with gmsh and %s.geo file.\n'%geofile)
    print('               -laplacian(u)+<c,grad(u)>+a0*u = g\n')
    print(' with ')
    print('   Fmap = %s'%func2str(Fmap,source=False))
    print('   a0 = %s '%func2str(a0,source=False))
    for i in range(len(c)):
      print('   c[%d] = %s'%(i,str(c[i])))
    print('   u = %s'%func2str(u,source=False))
    print('   -laplacian(u) = %s'%func2str(f,source=False))
    print('   gradSu = %s'%func2str(gradSu,source=False))
    slib.print_separator()
    if verbose==10:
      return

  print('*** Building the mesh using GMSH')
  meshfile=gmsh.buildmesh3ds(geodir+os.sep+geofile+'.geo',N, force=False,verbose=0)
  Th=siMesh(meshfile,trans=Fmap)
  print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
  print('*** Setting the surface BVP')
  cgradSu=lambda x,y,z: np.sum(np.array(gradSu(x,y,z))*np.array([c[0](x,y,z),c[1](x,y,z),c[2](x,y,z)]),axis=0)
  RHSfun=lambda x,y,z: f(x,y,z)+cgradSu(x,y,z)+a0(x,y,z)*u(x,y,z)

  Lop=Loperator(dim=3,d=2,A=[[1,None,None],[None,1,None],[None,None,1]],c=c,a0=a0)
  if is_function(a0):
    RHSfun=lambda x,y,z: f(x,y,z)+cgradSu(x,y,z)+a0(x,y,z)*u(x,y,z)
  else:
    RHSfun=lambda x,y,z: f(x,y,z)+cgradSu(x,y,z)+a0*u(x,y,z)
  pde=PDE(Op=Lop,f=RHSfun)
  bvp=BVP(Th,pde=pde)
  for lab in [1,3]:
    bvp.setDirichlet(lab,u)
  gN=lambda x,y,z: -gradSu(x,y,z)[2]
  gR=lambda x,y,z: -gradSu(x,y,z)[2]+alpha(x,y,z)*u(x,y,z)
  bvp.setRobin(2,gN)
  bvp.setRobin(4,gR,ar=alpha)  
  print('*** Solving the surface BVP')
  U=bvp.solve()
  
  if graphics and isModuleFound('fc_simesh_mayavi'):
    if verbose>0:
      print('*** Plotting results')
    from mayavi import mlab
    import fc_simesh_mayavi.siMesh as simlab
    from fc_vfemp1_surface.mayavi import plotSBVPsolution
    plotSBVPsolution(Th,U,uex=u,**kwargs)
  
  return {'bvp':bvp,'N':N,'uex':u,'a0':a0,'c':c,'f':f,'Fmap':Fmap,'gradSu':gradSu,'U':U}
  
def order(**kwargs):
  """
  Prints and plots orders of solutions of the BVP 
  
     `order()`
     `order(LN=[25,50,75,100])`
     `order(LN=[25,50,75,100],regular=False)`
            
  Optional (key/value) parameters
  -------------------------------
  `LN` : array or list of integers
      mesh size grows with `N in LN`.
  `graphics`: boolean (True by default)
      
  Others options are those of 
  * the `run` function in this file,
  * the `solve` function of class `BVP`:
    `solver` : to specify a solver for a sparse system A*x=b.
        Default is `solver=lambda A,b: scipy.sparse.linalg.spsolve(A,b)`
      
    `perm`   : to specify permutation in the sparse matrix. For example
        `perm=lambda A: scipy.sparse.csgraph.reverse_cuthill_mckee(A)`
  """  
  graphics=kwargs.get('graphics',True)
  res_order=slib.generic_order(run,**kwargs)
  if graphics and isModuleFound('fc_simesh_matplotlib'):
    import matplotlib.pyplot as plt
    from fc_tools.matplotlib import DisplayFigures
    from fc_vfemp1.matplotlib import plot_order
    plt.close('all')
    plt.ion()
    plt.figure()
    plot_order(res_order)  