import numpy as np
import os

from fc_tools.others import func2str,isModuleFound,is_function
from fc_oogmsh import gmsh
from fc_simesh.siMesh import siMesh
from fc_vfemp1.operators import Loperator
from fc_vfemp1.BVP import BVP,PDE

from fc_vfemp1_surface.sys import get_geo
import fc_vfemp1_surface.lib as slib
from fc_vfemp1_surface.examples import sBVPsamples


def run(**kwargs):
  N=kwargs.pop('N',30)
  verbose=kwargs.pop('verbose',1)
  u=kwargs.pop('u',sBVPsamples.uxyz01)
  f=kwargs.pop('f',sBVPsamples.fxyz01)
  Fmap=kwargs.pop('Fmap',sBVPsamples.Fmapxyz01)
  a0=kwargs.pop('a0',1)
  graphics=kwargs.pop('graphics',1)
  geofile='sphere4surf'
  (geodir,geofile)=get_geo(3,2,geofile)
  geoFile=geodir+os.sep+geofile+'.geo'
  if verbose>0:
    slib.print_separator()
    slib.PrintCopyright()
    slib.print_separator()
    print('      Surface BVP on unit sphere maps by function Fmap')
    print('      Mesh generated with gmsh and %s.geo file.\n'%geofile)
    print('               -laplacian(u)+a0*u = f\n')
    print(' with ')
    print('   Fmap = %s'%func2str(Fmap,source=False))
    print('   a0 = %s '%func2str(a0,source=False))
    print('   f = %s'%func2str(f,source=False))
    print('   u = %s'%func2str(u,source=False))
    slib.print_separator()
  print('*** Building the mesh using GMSH')
  meshfile=gmsh.buildmesh3ds(geodir+os.sep+geofile+'.geo',N, force=False,verbose=0)
  Th=siMesh(meshfile,trans=Fmap)
  print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
  print('*** Setting the surface BVP')
  Lop=Loperator(dim=3,d=2,A=[[1,None,None],[None,1,None],[None,None,1]],a0=a0)
  if is_function(a0):
    RHSf=lambda x,y,z: f(x,y,z)+a0(x,y,z)*u(x,y,z)
  else:
    RHSf=lambda x,y,z: f(x,y,z)+a0*u(x,y,z)
  pde=PDE(Op=Lop,f=RHSf)
  bvp=BVP(Th,pde=pde)
  print('*** Solving the surface BVP')
  U=bvp.solve()#local=True)
  
  if graphics and isModuleFound('fc_simesh_mayavi'):
    if verbose>0:
      print('*** Plotting results')
    from mayavi import mlab
    import fc_simesh_mayavi.siMesh as simlab
    from fc_vfemp1_surface.mayavi import plotSBVPsolution
    plotSBVPsolution(Th,U,uex=u,**kwargs)
  
  return {'bvp':bvp,'N':N,'uex':u,'a0':a0,'f':f,'Fmap':Fmap,'U':U}
  
def order(**kwargs):
  """
  Prints and plots orders of solutions of the BVP 
  
     `order()`
     `order(LN=[25,50,75,100])`
     `order(LN=[25,50,75,100],regular=False)`
            
  Optional (key/value) parameters
  -------------------------------
  `LN` : array or list of integers
      mesh size grows with `N in LN`.
  `graphics`: boolean (True by default)
      
  Others options are those of 
  * the `run` function in this file,
  * the `solve` function of class `BVP`:
    `solver` : to specify a solver for a sparse system A*x=b.
        Default is `solver=lambda A,b: scipy.sparse.linalg.spsolve(A,b)`
      
    `perm`   : to specify permutation in the sparse matrix. For example
        `perm=lambda A: scipy.sparse.csgraph.reverse_cuthill_mckee(A)`
  """  
  graphics=kwargs.get('graphics',True)
  res_order=slib.generic_order(run,**kwargs)
  if graphics and isModuleFound('fc_simesh_matplotlib'):
    import matplotlib.pyplot as plt
    from fc_tools.matplotlib import DisplayFigures
    from fc_vfemp1.matplotlib import plot_order
    plt.close('all')
    plt.ion()
    plt.figure()
    plot_order(res_order)  